import enum
import logging
import os
from pathlib import Path

import typer
from rich.logging import RichHandler
from rich.pretty import pprint
from typer.core import TyperGroup

from .cli.folder import app as folder_app
from .cli.run import monitor as monitor_cmd
from .cli.run import tidy as tidy_cmd
from .cli.sieve_export import sieve_export
from .config_state import state
from .utils import config_file_typer_option


class OrderCommands(TyperGroup):
    def list_commands(self, ctx: typer.Context):
        """Return list of commands in the order they appear."""
        return list(self.commands)  # get commands using self.commands


class LogLevel(enum.Enum):
    DEBUG = "debug"
    INFO = "info"
    WARNING = "warning"
    ERROR = "error"


def main_callback(log_level: LogLevel = typer.Option("info", case_sensitive=False)):
    if os.isatty(0):
        handler = RichHandler(
            rich_tracebacks=True, show_time=False, show_path=(log_level == LogLevel.DEBUG)
        )
    else:
        handler = logging.StreamHandler()
        formatter = logging.Formatter("%(levelname)s - %(message)s")
        handler.setFormatter(formatter)

    logging.basicConfig(
        level=log_level.name,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[handler],
    )


app = typer.Typer(no_args_is_help=True, cls=OrderCommands, callback=main_callback)


app.command(rich_help_panel="Main commands")(monitor_cmd)
app.command(rich_help_panel="Main commands")(tidy_cmd)
app.command(rich_help_panel="Main commands")(sieve_export)


@app.command(rich_help_panel="Util commands")
def check(config_file: Path = config_file_typer_option):
    """Load the config and connect to the IMAP server."""
    state.load_config_file(config_file)
    state.create_mailbox()
    pprint(state.rules)


app.add_typer(folder_app, rich_help_panel="Util commands")

if __name__ == "__main__":
    app()



import logging
import os
import ssl
import sys
from pathlib import Path
from typing import List

import certifi
import yaml
from imap_tools import BaseMailBox, MailBox

from mmuxer.models.action import Action, ActionLoader, DeleteAction, FlagAction, MoveAction
from mmuxer.models.enums import Flag
from mmuxer.models.rule import Rule
from mmuxer.models.script import PythonScript
from mmuxer.models.settings import Settings
from mmuxer.utils import ParseException

logger = logging.getLogger(__name__)


def make_ssl_context(ssl_ciphers):
    context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    if ssl_ciphers is not None:
        context.set_ciphers(ssl_ciphers)
    context.load_verify_locations(
        cafile=os.path.relpath(certifi.where()), capath=None, cadata=None
    )
    return context


default_actions = {
    "delete": DeleteAction(),
    "trash": MoveAction(dest="Trash"),
    "mark_read": FlagAction(flag=Flag.SEEN),
}


class State:
    __slots__ = ("_settings", "_rules", "_scripts", "_mailbox", "_config_file", "actions")

    def __init__(self):
        self._settings = None
        self._rules = None
        self._scripts = None
        self._mailbox = None
        self._config_file = None
        self.actions: dict[str, Action] = default_actions

    def load_config_file(self, config_file: Path):
        logger.info("Loading config from %s", config_file)
        self._config_file = config_file
        self._parse_config_file()

    def reload_config_file(self):
        current_state = {key: getattr(self, key) for key in self.__slots__}
        try:
            self._parse_config_file()
        except Exception:
            logger.error("Failed loading new configuration, reverting to previous configuration.")
            for key, value in current_state.items():
                setattr(self, key, value)
            return
        logger.info("Succesfuly parsed new configuration.")

    def create_mailbox(self):
        logger.debug(f"Connecting to {self.settings.server} with {self.settings.username}")
        ssl_context = make_ssl_context(self.settings.ssl_ciphers)
        self._mailbox = MailBox(self.settings.server, ssl_context=ssl_context).login(
            self.settings.username, self.settings.password
        )
        logger.info(f"Connected to {self.settings.server} with {self.settings.username}")

    def _parse_config_file(self):
        config_raw = self.config_file.read_text()
        try:
            config_dict = yaml.safe_load(config_raw)
        except Exception:
            logger.error("The config file is not valid yaml, check that the yaml syntax is valid.")
            exit(1)

        if not isinstance(config_dict, dict) or not "settings" in config_dict:
            logger.error("The config should be mapping. 'settings' is a required key")
            exit(1)

        all_keys = {"rules", "actions", "scripts", "settings"}
        extra_keys = config_dict.keys() - all_keys
        if extra_keys:
            logger.warning(f"The following keys in the config are not recognized: {extra_keys}")

        try:
            self._settings = Settings.parse_data(config_dict["settings"])
        except ParseException as exc:
            logger.error(exc.format("the 'settings' section of the configuration"))

            sys.exit(1)

        parsed_rules = []
        for rule_dict in config_dict["rules"]:
            try:
                parsed_rules.append(Rule.parse_data(rule_dict))
            except ParseException as exc:
                logger.error(exc.format("the following rules entry"))
                sys.exit(1)
        self._rules = parsed_rules

        if "actions" in config_dict:
            if not isinstance(config_dict["actions"], dict):
                logger.error(f"'actions' value is not a mapping: {config_dict['actions']}")
                sys.exit(1)
            for action_name, action_dict in config_dict["actions"].items():
                if action_name in self.actions:
                    logger.warning(f"Overriding default action f{action_name}")
                try:
                    self.actions[action_name] = ActionLoader.parse_data(action_dict).__root__
                except ParseException as exc:
                    logger.error(exc.format("the following action entry"))
                    sys.exit(1)
            self._rules = parsed_rules

        self._scripts = []
        for script_data in config_dict.get("scripts", []):
            try:
                self._scripts.append(PythonScript.parse_data(script_data))
            except ParseException as exc:
                logger.error(exc.format("the following script entry"))
                exit(1)

    @property
    def config_file(self) -> Path:
        if self._config_file is None:
            raise Exception("Uninitialized config_file")
        return self._config_file

    @property
    def settings(self) -> Settings:
        if self._settings is None:
            raise Exception("Uninitialized settings")
        return self._settings

    @property
    def rules(self) -> List[Rule]:
        if self._rules is None:
            raise Exception("Uninitialized rules")
        return self._rules

    @property
    def mailbox(self) -> BaseMailBox:
        if self._mailbox is None:
            raise Exception("Uninitialized mailbox")
        return self._mailbox

    @property
    def scripts(self) -> List[PythonScript]:
        if self._scripts is None:
            raise Exception("Uninitialized rules")
        return self._scripts


state = State()



from pydantic import BaseModel as PydanticBaseModel
from pydantic import ValidationError

from mmuxer.utils import ParseException


class BaseModel(PydanticBaseModel):
    class Config:
        extra = "forbid"

    @classmethod
    def parse_data(cls, data):
        try:
            return cls.parse_obj(data)
        except ValidationError as exc:
            raise ParseException.from_validation_error(
                exc=exc,
                full_content=data,
            )



import logging
from abc import abstractmethod
from typing import Literal, Union

from imap_tools import BaseMailBox, MailMessage

from mmuxer.utils import format_message

from .common import BaseModel
from .enums import Flag

logger = logging.getLogger(__name__)


class BaseAction(BaseModel):
    @abstractmethod
    def _apply(self, mailbox: BaseMailBox, message: MailMessage) -> None:
        """Implementation of the action"""
        pass

    @abstractmethod
    def format(self, message) -> str:
        """Used to print information in terminal"""
        pass

    def skip(self, message) -> bool:
        """Custom function to skip action in certain conditions"""
        return False

    def apply(self, mailbox: BaseMailBox, message: MailMessage, dry_run: bool):
        if self.skip(message):
            logger.debug("SKIP %s", self.format(message))
            return
        if dry_run:
            logger.info("%s [DRY_RUN]", self.format(message))
        else:
            logger.info(self.format(message))
            self._apply(mailbox, message)

    @abstractmethod
    def to_sieve(self) -> str:
        """Used to render sieve files"""


class MoveAction(BaseAction):
    action: Literal["move"] = "move"
    dest: str

    def skip(self, message):
        return getattr(message, "associated_folder", None) == self.dest

    def _apply(self, mailbox: BaseMailBox, message: MailMessage):
        mailbox.move(message.uid, self.dest)

    def format(self, message):
        return f"MOVE {format_message(message)} --> {self.dest}"

    def to_sieve(self):
        return f'fileinto "{self.dest}"'


class DeleteAction(BaseAction):
    action: Literal["delete"] = "delete"

    def _apply(self, mailbox: BaseMailBox, message: MailMessage):
        mailbox.delete([message.uid])

    def format(self, message):
        return f"DELETE {format_message(message)}"

    def to_sieve(self):
        return "discard"


class FlagAction(BaseAction):
    action: Literal["flag"] = "flag"
    flag: Flag

    def _apply(self, mailbox: BaseMailBox, message: MailMessage):
        mailbox.flag([message.uid], flag_set={self.flag.imap}, value=True)

    def format(self, message):
        return f"FLAG {format_message(message)} {self.flag.name}"

    def to_sieve(self):
        return f'setflag "{self.flag.sieve}"'


class UnflagAction(BaseAction):
    action: Literal["flag"] = "flag"
    flag: Flag

    def _apply(self, mailbox: BaseMailBox, message: MailMessage):
        mailbox.flag([message.uid], flag_set={self.flag.imap}, value=False)

    def format(self, message):
        return f"UNFLAG {format_message(message)} {self.flag.name}"

    def to_sieve(self):
        return f'removeflag "{self.flag.sieve}"'


Action = Union[MoveAction, DeleteAction, FlagAction, UnflagAction]


class ActionLoader(BaseModel):
    __root__: Action



import imaplib
import logging
import time
from threading import Event, Thread

from imap_tools import AND
from watchfiles import watch

from .config_state import state
from .models.rule import apply_list

logger = logging.getLogger(__name__)

flag = Event()


class MonitorWorker(Thread):
    def __init__(self, dry_run, folder):
        self.dry_run = dry_run
        self.folder = folder
        super().__init__()

    def run(self):
        logger.debug("Starting IMAP monitor thread.")
        box = state.mailbox
        while True:
            if flag.isSet():
                logger.info("Config change detected, reloading")
                state.reload_config_file()
                flag.clear()
            try:
                responses = box.idle.wait(timeout=state.settings.imap_wait_timeout)
                if responses:
                    for msg in box.fetch(AND(seen=False), mark_seen=False):
                        msg.associated_folder = self.folder
                        logger.info(
                            f"Found message [{{{msg.uid}}} {msg.from_} -> {msg.to} '{msg.subject}']"
                        )
                        apply_list(state.rules, box, msg, self.dry_run)
                        for script in state.scripts:
                            script.apply(msg, dry_run=self.dry_run)
            except imaplib.IMAP4.abort:
                logger.warning("IMAP connection aborted, reconnecting ")
                state.create_mailbox()
                box = state.mailbox
            except Exception:
                logger.exception("An error occured, reconnecting...")
                state.create_mailbox()
                box = state.mailbox

            time.sleep(0.1)


class WatcherWorker(Thread):
    def run(self):
        logger.debug("Starting file watcher thread, monitoring %s.", state.config_file)
        # force polling because of a bug with vim.
        # The other solution would be to watch the whole folder
        for _change in watch(
            state.config_file,
            force_polling=True,
            poll_delay_ms=1000,
        ):
            flag.set()



import logging
from collections import Counter

import typer
import yaml
from imap_tools import MailMessage
from pydantic import ValidationError

config_file_typer_option = typer.Option(
    ..., exists=True, dir_okay=False, readable=True, resolve_path=True
)
logger = logging.getLogger(__name__)


def find_likely_error_location_and_message(exc: ValidationError):
    locations = [error["loc"] for error in exc.errors()]

    prefixes = [location[0 : i + 1] for location in locations for i in range(len(location))]
    longest_prefix = sorted((prefix for prefix in set(prefixes)), key=len)[-1]
    most_common_prefixes = Counter(prefixes).most_common()
    if most_common_prefixes[0][1] == 1:
        longest_prefix = longest_prefix

    else:
        longest_prefix = ()
        for prefix, count in Counter(prefixes).most_common():
            if count == 1:
                break
            if len(prefix) > len(longest_prefix):
                longest_prefix = prefix

    errors_with_loc = [error for error in exc.errors() if error["loc"] == longest_prefix]
    if len(errors_with_loc) == 1:
        message = errors_with_loc[0]["msg"]
    else:
        message = "could not parse"
    return longest_prefix, message


def get_from_keys_tuple(data, keys):
    if len(keys) == 0:
        return data
    key, *remaining_keys = keys
    if isinstance(data, dict) and not key in data:
        return "Missing field"
    return get_from_keys_tuple(data[key], remaining_keys)


def format_data(data):
    data_str = yaml.dump(data)
    return "╭\n" + "\n".join("│ " + line for line in data_str.splitlines()) + "\n╰"


class ParseException(Exception):
    def __init__(self, error_loc, bad_content, full_content, validation_error, message):
        self.error_loc = error_loc
        self.bad_content = bad_content
        self.full_content = full_content
        self.validation_error = validation_error
        self.message = message

    def format(self, desc_str):
        return (
            f"Failed parsing {desc_str}:\n"
            + format_data(self.full_content)
            + f"\nThe malformed configuration may come from '{self.message}' in\n"
            + format_data(self.bad_content)
        )

    @classmethod
    def from_validation_error(cls, exc: ValidationError, full_content):
        error_loc, message = find_likely_error_location_and_message(exc)
        if len(error_loc) == 0:
            bad_content = full_content
        elif isinstance(error_loc[-1], str):
            bad_content = {error_loc[-1]: get_from_keys_tuple(full_content, error_loc)}
        else:
            bad_content = get_from_keys_tuple(full_content, error_loc)

        return ParseException(
            error_loc=error_loc,
            bad_content=bad_content,
            full_content=full_content,
            validation_error=exc,
            message=message,
        )


def format_message(msg: MailMessage):
    return f"[{{{msg.uid}}} {msg.from_} -> {msg.to} '{msg.subject}']"



import importlib
import importlib.util
import logging
import sys
from typing import Any, Dict

from imap_tools import MailMessage
from pydantic import PrivateAttr

from mmuxer.utils import format_message

from .common import BaseModel
from .condition import Condition

logger = logging.getLogger(__name__)


# https://stackoverflow.com/questions/67631/how-can-i-import-a-module-dynamically-given-the-full-path
class PythonScript(BaseModel):
    name: str
    script_path: str
    entrypoint: str
    condition: Condition
    kwargs: Dict[str, Any] = {}
    _callable: Any = PrivateAttr()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            self._callable = self._load_callable()
        except Exception:
            logger.exception(f"Failed loading code for script {self.name}")
            sys.exit(1)

    def apply(self, message: MailMessage, dry_run):
        if self.condition.eval(message):
            log_message = f"Executing script {self.name} for {format_message(message)}"
            if dry_run:
                logger.info("%s [DRY_RUN]", log_message)
            else:
                logger.info(log_message)
                try:
                    self._callable(message, **self.kwargs, logger=logger)
                except Exception:
                    logger.exception(f"Script {self.name} execution failed.")

    def _load_callable(self):
        spec = importlib.util.spec_from_file_location(self.name, self.script_path)
        if spec is None:
            raise Exception(f"Could not find python module to load at '{self.script_path}")

        script_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(script_module)

        entrypoint = getattr(script_module, self.entrypoint, None)
        if entrypoint is None:
            raise Exception(
                f"Could not find a function '{self.entrypoint}' in '{self.script_path}'."
            )
        return entrypoint



import logging
from pathlib import Path
from typing import Optional

import typer

from mmuxer.config_state import state
from mmuxer.utils import config_file_typer_option

logger = logging.getLogger(__name__)


def sieve_export(
    config_file: Path = config_file_typer_option,
    dest_file: Optional[Path] = typer.Option(
        None, help="File to write. If not provided, write to stdout."
    ),
):
    """Convert the rules of the give config file to sieve format."""
    state.load_config_file(config_file)
    rules_str = "\n".join(sieve_rule for rule in state.rules for sieve_rule in rule.to_sieve())
    output = f"""require ["fileinto","imap4flags","regex"];
{rules_str}
    """
    if dest_file is None:
        print(output)
    else:
        dest_file.write_text(output)
        logger.info(f"Rules written to '{dest_file.resolve()}'")



from typing import Optional

from pydantic import BaseSettings

from mmuxer.models.common import BaseModel


def in_container():
    """Returns: True iff running in a container"""
    with open("/proc/1/cgroup") as ifh:
        value = ifh.read()
        return "docker" in value or "kubepod" in value


class BaseConfig:
    env_file = ".env"
    secrets_dir: Optional[str] = None


if in_container():
    BaseConfig.secrets_dir = "/run/secrets"


class Settings(BaseModel, BaseSettings):
    server: str
    username: str
    password: str
    ssl_ciphers: Optional[str] = None
    imap_wait_timeout: int = 60

    class Config(BaseConfig):
        pass



import logging
from pathlib import Path

import typer

from mmuxer.config_state import state
from mmuxer.models.rule import apply_list
from mmuxer.utils import config_file_typer_option
from mmuxer.workers import MonitorWorker, WatcherWorker

logger = logging.getLogger(__name__)


def tidy(
    config_file: Path = config_file_typer_option,
    folder: str = typer.Option(None, help="Folder to fetch the messages from"),
    dry_run: bool = typer.Option(False, help="Print actions instead of running them"),
):
    """Run once, on all messages of the INBOX (or the given folder)."""
    state.load_config_file(config_file)
    state.create_mailbox()
    box = state.mailbox
    if folder is not None:
        box.folder.set(folder)
    counter = 0
    for msg in box.fetch(bulk=True):
        msg.associated_folder = folder
        apply_list(state.rules, box, msg, dry_run)
        for script in state.scripts:
            script.apply(msg, dry_run=dry_run)
        counter += 1
    print()
    print(f"{counter} messages parsed.")


def monitor(
    config_file: Path = config_file_typer_option,
    folder: str = typer.Option(None, help="Folder to fetch the messages from"),
    dry_run: bool = typer.Option(False, help="Print actions instead of running them"),
    auto_reload: bool = typer.Option(
        False, help="Auto-reload config file on modification (EXPERIMENTAL)"
    ),
):
    """Monitor mailbox, and apply rules on unseen messages."""
    state.load_config_file(config_file)
    state.create_mailbox()

    if auto_reload:
        MonitorWorker(folder, dry_run).start()
        WatcherWorker().start()
    else:
        MonitorWorker(folder, dry_run).run()
